/***************************************************************************
                           chttp.cpp  -  description
                             -------------------
    begin                : Mon Jul 29 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>

#include <dclib/dcos.h>
#include <dclib/core/clist.h>
#include <dclib/core/ccallback.h>
#include <dclib/core/cmanager.h>
#include <dclib/dcobject.h>

#include "chttp.h"

/** */
CHttp::CHttp()
{
	m_pHttpCallback = 0;
	m_pCallback     = 0;

	m_pMessageList = new CList<CObject>();
}

/** */
CHttp::~CHttp()
{
	if ( m_pCallback )
	{
		CManager::Instance()->Remove( m_pCallback );
		delete m_pCallback;
		m_pCallback = 0;
	}

	Disconnect();

	if ( m_pHttpCallback )
	{
		delete m_pHttpCallback;
		m_pHttpCallback = 0;
	}

	delete m_pMessageList;
}

/** */
int CHttp::Callback( CObject *, CObject * )
{
	Thread(0);
	return 0;
}

/** */
int CHttp::GetUrl( CString url )
{
	int i,i1;

	CString s(url);
	CString sRealHost;
	CString sRealPort;

	//Disconnect();

	iErrorCode = 0;
	bData      = FALSE;
	sHeader    = "";
	baData.SetSize(0);

	if ( s == "" )
	{
		return -1;
	}

	CString header = s.Mid(0,7);

	header = header.ToUpper();

	if ( header == "HTTP://" )
	{
		s = s.Mid(7,s.Length()-7);
	}

	if ( (i = s.Find(':')) >= 0 )
	{
		sHost = s.Mid(0,i);
	}

	if ( (i1 = s.Find('/')) < 0 )
	{
		return -1;
	}

	if ( i >= 0 )
	{
		sPort = s.Mid(i+1,i1-i-1);
	}
	else
	{
		sHost = s.Mid(0,i1);
		sPort = "80";
	}

	char *proxy = getenv("http_proxy");

	if( proxy )
	{
		CString sProxyTemp ( proxy );
		CString header = sProxyTemp.Mid(0,7);

		header = header.ToUpper();

		if ( header == "HTTP://" )
		{
			sProxyTemp = sProxyTemp.Mid(7,sProxyTemp.Length()-7);
		}

		if ( (i = sProxyTemp.Find(':')) >= 0 )
		{
			sProxy = sProxyTemp.Mid(0,i);
		}

		if ( (i1 = sProxyTemp.Find('/')) < 0 )
		{
			return -1;
		}

		if ( i >= 0 )
		{
			sProxyPort = sProxyTemp.Mid(i+1,i1-i-1);
		}
		else
		{
			sProxy = sProxyTemp.Mid(0,i1);
			sProxyPort = "8080";
		}

		sUrl = url;
		sRealHost = sProxy;
		sRealPort = sProxyPort;
	}
	else
	{
		sUrl = s.Mid(i1,s.Length()-i1);
		sProxy = "";
		sRealHost = sHost;
		sRealPort = sPort;
	}

	if ( sProxy != "" )
	{
		printf("PROXY: '%s:%s'\n",sProxy.Data(),sProxyPort.Data());
	}

	printf("HOST : '%s:%s'\n",sHost.Data(),sPort.Data());
	printf("URL  : '%s'\n",sUrl.Data());

	if ( Connect( sRealHost, sRealPort.asINT() ) == ecsERROR )
	{
		return -1;
	}

	if ( !m_pCallback )
	{
		m_pCallback = new CCallback<CHttp>( this, &CHttp::Callback );
		CManager::Instance()->Add( m_pCallback );
	}

	return 0;
}

/** */
bool CHttp::GetData( CByteArray * ba )
{
	bool err = FALSE;

	if ( (iErrorCode == 200) && (bData == TRUE) && (eMode == estNONE) )
	{
		if ( (nContentLength == -1) || (nContentLength == baData.Size()) )
		{
			ba->SetSize(0);
			ba->Append( baData.Data(), baData.Size() );
			err = TRUE;
		}
	}

	return err;
}

/** */
int CHttp::CallBack_SendObject( CObject * Object )
{
	int err;

	if ( m_pHttpCallback )
	{
		err = m_pHttpCallback->notify(this,Object);
	}
	else
	{
		err = DC_CallBack( Object );
	}

	if ( err == -1 )
	{
		printf("CallBack failed (state)...\n");
		delete Object;
	}

	return err;
}

/** */
void CHttp::ConnectionState( eConnectionState state )
{
	CMessageConnectionState *Object;

	Object = new CMessageConnectionState();

	Object->m_eState   = state;
	Object->m_sMessage = GetSocketError();

	if ( state == estCONNECTED )
	{
		eMode = estTRANSFERHANDSHAKE;
	}
	else if ( state == estDISCONNECTED )
	{
		eMode = estNONE;
	}

	// callback over notify because connection class is locked
	m_pMessageList->Add(Object);
}

/** */
void CHttp::DataAvailable( const char * buffer, int len )
{
	int i,i1,i2;

	if ( eMode == estTRANSFERDOWNLOAD )
	{
		if ( bData == FALSE )
		{
			for(i=0;i<len;i++)
			{
				if ((sHeader.Length() > 4) &&
				    (sHeader.Data()[sHeader.Length()-4] == 0x0d) &&
				    (sHeader.Data()[sHeader.Length()-3] == 0x0a) &&
				    (sHeader.Data()[sHeader.Length()-2] == 0x0d) &&
				    (sHeader.Data()[sHeader.Length()-1] == 0x0a) )
				{
					//sHeader += CString().Set(buffer,i);
					//i+=4;

					// parse header ...
					if ( (i1 = sHeader.Find("HTTP/1.0 ")) == -1 )
					{
						if ( (i1 = sHeader.Find("HTTP/1.1 ")) == -1 )
						{
							printf("wrong proto '%s'\n",sHeader.Data());
							Disconnect();
							return;
						}
					}

					if ( (i2 = sHeader.Find(' ',i1+9)) == -1 )
					{
						printf("wrong proto '%s'\n",sHeader.Data());
						Disconnect();
					}
					else
					{
						iErrorCode = sHeader.Mid(i1+9,i2-i1-9).asINT();

						if ( iErrorCode == 200 )
						{
							printf("no error\n");
							bData = TRUE;
						}
						else
						{
							printf("http error %d\n",iErrorCode);
							Disconnect();
						}
					}

					nContentLength = -1;

					if ( (i1 = sHeader.Find("Content-Length: ")) != -1 )
					{
						if ( (i2 = sHeader.Find(0x0d,i1)) != -1 )
						{
							nContentLength = sHeader.Mid(i1+16,i2-i1-16).asLONG();
						}
					}

					break;
				}
				else
				{
					sHeader += buffer[i];
				}
			}

			if ( (bData == FALSE) && (iErrorCode == 0) )
			{
				//sHeader += buffer;
			}
			else if ( iErrorCode != 0 )
			{
				AppendData( buffer+i,len-i );
			}
		}
		else
		{
			AppendData( buffer, len );
		}

		//printf("DATA: %ld\n",baData.Size());
	}
}

/** */
void CHttp::AppendData( const char * buffer, int len )
{
	CMessageTransfer  * msg = new CMessageTransfer();

	if ( nContentLength != -1 )
	{
		msg->m_nLength = nContentLength;
	}

	msg->m_nTransfered = baData.Size();

	m_pMessageList->Add(msg);

	baData.Append((const unsigned char*)buffer,len);
}

/** */
void CHttp::DataSend()
{
	CString header;

	if ( eMode == estTRANSFERHANDSHAKE )
	{
		// send get request
		header  = "GET ";
		header += sUrl;
		header += " HTTP/1.1";
		header += "\xd\xa";
		header += "User-Agent: DCGUI v";
		header += VERSION;
		header += "\xd\xa";
		header += "Referer: http://dc.ketelhot.de";
		header += "\xd\xa";
		header += "Host: ";
		header += sHost;
		header += "\xd\xa";
		header += "\xd\xa";

		if ( Write( (const unsigned char*)header.Data(), header.Length() ) != 0 )
		{
			eMode = estNONE;
		}
		else
		{
			eMode = estTRANSFERDOWNLOAD;
		}
	}
}

/** */
void CHttp::DataTimeout()
{
	Disconnect(TRUE);
	printf("data timeout\n");
}

/** */
void CHttp::Notify()
{
	CObject * object = 0;

	while ( (object=m_pMessageList->Next(0)) != 0 )
	{
		m_pMessageList->Remove(object);
		CallBack_SendObject(object);
	}
}

/** */
CString CHttp::Encode( CString s )
{
	CString s1;

	s1 = s.Replace("&","&amp;");
	s1 = s1.Replace(">","&gt;");
	s1 = s1.Replace("<","&lt;");
	s1 = s1.Replace("\x7f","&#x25a1;");
	s1 = s1.Replace("=","&#61;");
	s1 = s1.Replace(" ","&#32;");

	return s1;
}

/** */
CString CHttp::Decode( CString s )
{
	CString s1;

	s1 = s.Replace("&gt;",">");
	s1 = s1.Replace("&lt;","<");
	s1 = s1.Replace("&#x25a1;","\x7f");
	s1 = s1.Replace("&#61;","=");
	s1 = s1.Replace("&#32;"," ");
	s1 = s1.Replace("&amp;","&");

	return s1;
}
