/***************************************************************************
                          chubsearch.cpp  -  description
                             -------------------
    begin                : Fri Mar 15 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>

#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#endif

#include <dclib/dcos.h>
#include <dclib/dcobject.h>
#include <dclib/cconfig.h>
#include <dclib/cclient.h>
#include <dclib/core/cmanager.h>
#include <dclib/cmessagehandler.h>
#include <dclib/cservermanager.h>
#include <dclib/cfilemanager.h>

#include "chubsearch.h"

CHubSearch * _pHubSearch;

CHubSearch::CHubSearch()
{
	m_sBuffer        = "";
	m_pHubServerList = 0;
	m_pSearchList    = 0;
	m_ehSearchType   = ehstNONE;
	m_bEnableTag     = FALSE;

	m_pHubSearchClientList = new CList<CObject>();

	_pHubSearch = this;

	m_pCallback = new CCallback<CHubSearch>( this, &CHubSearch::DC_ManagerCallBack );
	CManager::Instance()->Add( m_pCallback );

	SetInstance(this);
}

CHubSearch::~CHubSearch()
{
	CManager::Instance()->Remove( m_pCallback );
	if ( m_pCallback )
	{
		delete m_pCallback;
		m_pCallback = 0;
	}

	// clear the client list
	HubSearchClientListThread.Lock();

	sHubSearchClient * HubSearchClient;

	if ( m_pHubSearchClientList != 0 )
	{
		while ( ( (HubSearchClient = (sHubSearchClient *)m_pHubSearchClientList->Next(0)) != NULL) )
		{
			HubSearchClient->m_pClient->SetCallBackFunction(0);
			HubSearchClient->m_pClient->Disconnect(TRUE);
			delete HubSearchClient->m_pClient;
			HubSearchClient->m_pClient = 0;
			m_pHubSearchClientList->Remove(HubSearchClient);
			delete HubSearchClient;
		}

		delete m_pHubSearchClientList;
		m_pHubSearchClientList = 0;
	}

	HubSearchClientListThread.UnLock();

	// clear hubserverlist
	if ( m_pHubServerList != 0 )
	{
		delete m_pHubServerList;
		m_pHubServerList = 0;
	}
}

/** */
void CHubSearch::ConnectionState( eConnectionState state )
{
	CMessageConnectionState *Object;

	Object = new CMessageConnectionState();

	Object->m_eState   = state;
	Object->m_sMessage = GetSocketError();

	if ( DC_CallBack( Object ) == -1 )
	{
		printf("CallBack failed (state)...\n");
		delete Object;
	}
}

/** */
int CHubSearch::HandleMessage( char * c, int )
{
	int p=0;
	eDCMessage type;
	CObject * Object;

	CMessageHandler MessageHandler;

	CString s;

	s = c;

	while( (type=MessageHandler.Parse(&s,p,Object=0)) != DC_MESSAGE_PARSE_ERROR )
	{
		switch (type)
		{
			case DC_MESSAGE_SEARCHRESULT:
			{
				break;
			}

			default:
			{
				// remove message
				delete Object;
				Object = 0;
			}
		}

		if (Object)
		{
			((CDCMessage*)Object)->m_eType = type;

			SendSearchResult(Object);
		}
	}

	return 0;
}

/** passive search results */
bool CHubSearch::HandleMessage( CMessageSearchResult * MessageSearchResult )
{
	bool res = FALSE;

	if ( m_ehSearchType == ehstEXTERNAL )
	{
		if ( CDownloadManager::Instance() )
		{
			res = CDownloadManager::Instance()->DLM_HandleSearch(MessageSearchResult);
		}
	}
	else
	{
		SendSearchResult(MessageSearchResult);
		res = TRUE;
	}

	return res;
}

/** */
void CHubSearch::DataAvailable( const char * buffer, int len )
{
	m_sBuffer = buffer;

	// add traffic control
	CSocket::m_Traffic.AddTraffic(ettCONTROLRX,len);

	// add '|' for parser
	m_sBuffer += '|';

	HandleMessage( (char*)m_sBuffer.Data(), m_sBuffer.Length() );
}

/** */
bool CHubSearch::ExternalSearch( bool enabled )
{
	bool res = FALSE;

	if ( (m_ehSearchType == ehstNONE) && (enabled) )
	{
		m_ehSearchType = ehstEXTERNAL;
		res = TRUE;
	}
	else if ( (m_ehSearchType == ehstEXTERNAL) && (!enabled) )
	{
		m_ehSearchType = ehstNONE;
		res = TRUE;
	}

	return res;
}

/** */
int CHubSearch::StartSearch( eHubSearchHubs hubs, bool multi, CList<CObject> * searchlist, CStringList * hublist, CString s )
{
	CObject * object;
	sHubSearchClient * HubSearchClient;

	m_bMultiSearch = multi;
	m_ehSearchType = ehstNONE;

	// reset current server
	m_sCurrentServer = 0;
	m_nCurrentHub    = 0;
	m_nError         = 0;

	// set start time
	m_tStartTime = time(0);

	// clear client list
	HubSearchClientListThread.Lock();

	while ( ( (HubSearchClient = (sHubSearchClient *)m_pHubSearchClientList->Next(0)) != NULL) )
	{
		m_pHubSearchClientList->Remove(HubSearchClient);

		HubSearchClient->m_pClient->SetCallBackFunction(0);
		HubSearchClient->m_pClient->Disconnect();
		delete HubSearchClient->m_pClient;
		HubSearchClient->m_pClient = 0;
		delete HubSearchClient;
	}

	HubSearchClientListThread.UnLock();

	if ( m_pHubServerList != 0 )
	{
		delete m_pHubServerList;
		m_pHubServerList = 0;
	}

	m_pSearchList = searchlist;

	// check if a user search in the list
	m_bHandleUserList = FALSE;
	object = 0;

	while ( (object=m_pSearchList->Next(object)) != 0 )
	{
		if ( ((CDCMessage*)object)->m_eType == DC_MESSAGE_SEARCH_USER )
		{
			m_bHandleUserList = TRUE;
			break;
		}
	}
	
	// connected hubs
	if ( (hubs == ehshCONNECTEDALL) ||
	     (hubs == ehshCONNECTEDSINGLE) )
	{
		// get connected hub count
		if ( CServerManager::Instance()->GetConnectedHubCount() == 0 )
		{
			return -1;
		}

		m_sHubName = s;
		m_pCurrentSearchObject = 0;
		m_tHubSearchTimeout = time(0);

		m_ehSearchType = ehstLOCAL;

		// send first search
		if ( SendSearch(m_sHubName) <= 0 )
		{
			m_ehSearchType = ehstNONE;
			return -1;
		}
	}
	else
	{
		// all public hubs
		if ( hubs == ehshPUBLIC )
		{
			// get public hub server list
			if ( hublist != 0 )
				m_pHubServerList = hublist;
			else
				m_pHubServerList = CServerManager::Instance()->GetPublicHubServerList();
		}
		// all bookmark hubs
		else if ( hubs == ehshBOOKMARK )
		{
			// get bookmark hub server list
			m_pHubServerList = CServerManager::Instance()->GetBookmarkHubServerList();
		}

		if ( (m_pHubServerList == 0) || (m_pHubServerList->Count() <= 0) )
		{
			return -1;
		}

		m_ehSearchType = ehstGLOBAL;
	}

	Start();

	return 0;
}

/** */
void CHubSearch::StopSearch()
{
	m_ehSearchType = ehstSTOP;

	Disconnect();
}

/** */
void CHubSearch::SendDebug( CString s )
{
	CMessageLog * MessageLog = new CMessageLog();

	MessageLog->sMessage = s;
	MessageLog->m_eType  = DC_MESSAGE_LOG;

	if ( DC_CallBack( MessageLog ) == -1 )
	{
		printf("CallBack failed (state)...\n");
		delete MessageLog;
	}
}

/** */
void CHubSearch::SendSearchResult( CObject * object )
{
	if ( DC_CallBack( object ) == -1 )
	{
		printf("CallBack failed (state)...\n");
		delete object;
	}
}

/** */
void CHubSearch::UpdateClients()
{
	if ( m_pHubSearchClientList != 0 )
	{
		sHubSearchClient * HubSearchClient = 0;

		while ( ( (HubSearchClient = (sHubSearchClient *)m_pHubSearchClientList->Next(HubSearchClient)) != NULL) )
		{
			HubSearchClient->m_pClient->Thread(0);
		}
	}
}

/** */
int CHubSearch::DC_ManagerCallBack( CObject *, CObject * )
{
	UpdateClients();

	if ( (_pHubSearch->m_ehSearchType == ehstWAITTIMEOUT) &&
	     ((time(0)-m_tHubSearchTimeout) >= 120) )
	{
		// stop search
		StopSearch();
	}
	else if ( (_pHubSearch->m_ehSearchType == ehstGLOBAL) ||
	     (_pHubSearch->m_ehSearchType == ehstWAITTIMEOUT) ||
	     (_pHubSearch->m_ehSearchType == ehstSTOP) )
	{
		_pHubSearch->CheckClient();

		if ( _pHubSearch->m_ehSearchType == ehstGLOBAL )
		{
			_pHubSearch->NewClient();
		}

		if ( _pHubSearch->m_ehSearchType == ehstSTOP )
		{
			_pHubSearch->HubSearchClientListThread.Lock();

			if ( _pHubSearch->m_pHubSearchClientList->Count() == 0 )
			{
				_pHubSearch->m_ehSearchType = ehstNONE;
				Stop();
			}

			_pHubSearch->HubSearchClientListThread.UnLock();
		}
	}
	else if ( _pHubSearch->m_ehSearchType == ehstLOCAL )
	{
		if ( (time(0)-m_tHubSearchTimeout) >= 11 )
		{
			// send first search
			if ( SendSearch(m_sHubName) <= 0 )
			{
				m_ehSearchType = ehstWAITTIMEOUT;
			}

			m_tHubSearchTimeout = time(0);
		}
	}

	return 0;
}

/** */
void CHubSearch::CheckClient()
{
	sHubSearchClient * HubSearchClient,* OldHubSearchClient;

	HubSearchClientListThread.Lock();

	if ( m_pHubSearchClientList != 0 )
	{
		HubSearchClient = OldHubSearchClient = 0;

		// check all hubs in the list
		while ( ( (HubSearchClient = (sHubSearchClient *)m_pHubSearchClientList->Next(HubSearchClient)) != NULL) )
		{
			if ( (m_ehSearchType == ehstSTOP) && (HubSearchClient->m_bRemove != TRUE) )
			{
				// set flag to remove all clients
				HubSearchClient->m_pClient->SetCallBackFunction(0);
				HubSearchClient->m_bRemove = TRUE;

				OldHubSearchClient = HubSearchClient;
			}
			// remove the client or if timeout
			else if ( (HubSearchClient->m_bRemove == TRUE) || ((time(0)-HubSearchClient->m_tHubTimeout) >= 300) )
			{
				// check remove timeout
				if ( (HubSearchClient->m_bRemove == TRUE) &&
				     ((time(0)-HubSearchClient->m_tHubTimeout) < 6) &&
				     (m_ehSearchType != ehstSTOP) )
				{
					// hold this client in the list
					OldHubSearchClient = HubSearchClient;
				}
				else
				{
					HubSearchClient->m_pClient->SetCallBackFunction(0);

					if ( HubSearchClient->m_pClient->GetConnectionState() != estNONE )
					{
						if ( HubSearchClient->m_pClient->GetConnectionState() != estDISCONNECTING )
						{
							HubSearchClient->m_pClient->Disconnect(TRUE);
						}
						OldHubSearchClient = HubSearchClient;
					}
					else
					{
						HubSearchClient->m_pClient->Stop();
						delete HubSearchClient->m_pClient;
						HubSearchClient->m_pClient = 0;
						m_pHubSearchClientList->Remove(HubSearchClient);
						delete HubSearchClient;
						HubSearchClient = OldHubSearchClient;
					}
				}
			}
			else
			{
				// send search
				if ( (HubSearchClient->m_bEnableSearch == TRUE) && ((time(0)-HubSearchClient->m_tSearchTimeout) >= 11) )
				{
					if ( SendSearch(HubSearchClient) == FALSE )
					{
						// no more searches, remove the client
						HubSearchClient->m_bRemove = TRUE;
					}
					else
					{
						// reset hubtimeout
						HubSearchClient->m_tHubTimeout = time(0);
					}
				}

				OldHubSearchClient = HubSearchClient;
			}
		}
	}

	HubSearchClientListThread.UnLock();
}

/** */
void CHubSearch::RemoveClient( CClient * Client )
{
	HubSearchClientListThread.Lock();

	sHubSearchClient * HubSearchClient;

	if ( m_pHubSearchClientList != 0 )
	{
		HubSearchClient = 0;

		while ( (HubSearchClient = (sHubSearchClient *)m_pHubSearchClientList->Next((CObject*)HubSearchClient)) != NULL )
		{
			if ( HubSearchClient->m_pClient == Client )
			{
				HubSearchClient->m_tHubTimeout = time(0);
				HubSearchClient->m_pClient->SetCallBackFunction(0);
				HubSearchClient->m_bRemove = TRUE;
				break;
			}
		}
	}

	HubSearchClientListThread.UnLock();
}

/** */
void CHubSearch::EnableSearch( CClient * Client )
{
	HubSearchClientListThread.Lock();

	sHubSearchClient * HubSearchClient;

	if ( m_pHubSearchClientList != 0 )
	{
		HubSearchClient = 0;

		while ( (HubSearchClient = (sHubSearchClient *)m_pHubSearchClientList->Next((CObject*)HubSearchClient)) != NULL )
		{
			if ( HubSearchClient->m_pClient == Client )
			{
				HubSearchClient->m_bEnableSearch = TRUE;
				break;
			}
		}
	}

	HubSearchClientListThread.UnLock();
}

/** */
void CHubSearch::NewClient()
{
	int l;
	CString server,addr,s;
	sHubSearchClient * HubSearchClient;

	if ( m_pHubServerList == 0 )
	{
		return;
	}

	// done ?
	if ( m_pHubServerList->Count() == m_nCurrentHub )
	{
		SendDebug("Serverlist done\n");

		m_ehSearchType = ehstWAITTIMEOUT;
		m_tHubSearchTimeout = time(0);

		return;
	}

	HubSearchClientListThread.Lock();

	l = m_pHubSearchClientList->Count();

	HubSearchClientListThread.UnLock();

	for(;(l<m_nMaxThreads)&&(m_nCurrentHub<m_pHubServerList->Count());l++)
	{
		CClient * Client = new CClient();

		HubSearchClient = new sHubSearchClient;

		HubSearchClient->m_pClient = Client;
		HubSearchClient->m_bRemove = FALSE;
		HubSearchClient->m_tHubTimeout = time(0);

		if ( CConfig::Instance()->GetMode() == ecmPASSIVE )
		{
			HubSearchClient->m_tHubTimeout += 30;
		}

		Client->SetNick(CConfig::Instance()->GetSearchNick());
		Client->SetComment(CConfig::Instance()->GetDescription(!m_bEnableTag));
		Client->SetConnectionType(CConfig::Instance()->GetSpeed());
		Client->SetEMail(CConfig::Instance()->GetEMail());
		Client->SetVersion(VERSION);
		Client->SetShareSize(CString().setNum(CFileManager::Instance()->GetShareSize()));
		Client->SetMode(CConfig::Instance()->GetMode());
		Client->SetHandleUserList(m_bHandleUserList);
		Client->SetHandleSearch(FALSE);
		Client->SetHandleMyinfo(FALSE);
		Client->SetHandleForceMove(FALSE);
		Client->SetHandleTransfer(FALSE);
		Client->SetCallBackFunction( new CCallback<CHubSearch>( this, &CHubSearch::DC_ClientCallBack ) );

		if ( m_pHubServerList->Next((CObject*&)m_sCurrentServer) != 0 )
		{
			server = *m_sCurrentServer;

			m_nCurrentHub++;

			SendDebug( "Search on: [" + CString().setNum(m_nCurrentHub) + "] " + server );

			HubSearchClientListThread.Lock();

			m_pHubSearchClientList->Add((CObject*)HubSearchClient);

			Client->Connect(server);
		}
		else
		{
			// end on error
			m_nCurrentHub = m_pHubServerList->Count();
		}

		HubSearchClientListThread.UnLock();
	}
}

/** */
CString CHubSearch::GetSearchString( CMessageSearchFile * msg, bool multi )
{
	CString s = "";

	if ( msg->m_bLocal == TRUE )
		s += "Hub:";
	s += msg->m_sSource;
	s += " ";

	if ( msg->m_bSizeLimit == FALSE )
		s += "F?";
	else
		s += "T?";

	if ( msg->m_bSizeAtMost == FALSE )
		s += "F?";
	else
		s += "T?";

	s += CString().setNum(msg->m_nSize) + "?";
	s += CString().setNum(msg->m_eFileType) + "?";
	s += msg->m_sString.Replace(' ',"$") + "|";

	// passive searches with multisearch
	if ( msg->m_bLocal == TRUE )
		s = "$Search " + s;
	else if ( !multi )
		s = "$Search " + s;
	else
		s = "$MultiSearch " + s;

	return s;
}

/** send the search to the clients */
int CHubSearch::SendSearch( CString s )
{
	int i = 0;
	CString search;
	CList<DCHubObject> list;
	DCHubObject * HubObject;

	if ( (m_pCurrentSearchObject=m_pSearchList->Next(m_pCurrentSearchObject)) != 0 )
	{
		switch(((CDCMessage*)m_pCurrentSearchObject)->m_eType)
		{
			case DC_MESSAGE_SEARCH_USER:
			{
				CMessageSearchUser * msg = (CMessageSearchUser*)m_pCurrentSearchObject;

				if ( CServerManager::Instance()->IsUserOnline( msg->m_sNick, s, &list ) == TRUE )
				{
					HubObject = 0;

					while ( (HubObject=list.Next(HubObject)) != 0 )
					{
						CMessageSearchResultUser * MessageSearchResultUser = new CMessageSearchResultUser();

						MessageSearchResultUser->m_eType    = DC_MESSAGE_SEARCHRESULT_USER;
						MessageSearchResultUser->m_sHubName = HubObject->m_sHubName;
						MessageSearchResultUser->m_sNick    = msg->m_sNick;

						SendSearchResult(MessageSearchResultUser);
					}

					list.Clear();
				}

				i = 1;

				break;
			}

			case DC_MESSAGE_SEARCH_FILE:
			{
				CMessageSearchFile * msg = (CMessageSearchFile*)m_pCurrentSearchObject;

				search = GetSearchString( msg, m_bMultiSearch );

				if ( search != "" )
					i = CServerManager::Instance()->SendStringToConnectedServers(search,s);
				break;
			}

			default:
			{
				break;
			}
		}
	}

	return i;
}

/** send the search to this client */
bool CHubSearch::SendSearch( sHubSearchClient * HubSearchClient )
{
	CString search;
	CObject *object;
	bool res = FALSE;

	object = HubSearchClient->m_pCurrentSearch;

	while ( (res == FALSE) && ((object=m_pSearchList->Next(object)) != 0) )
	{
		switch(((CDCMessage*)object)->m_eType)
		{
			case DC_MESSAGE_SEARCH_USER:
			{
				CMessageSearchUser * msg = (CMessageSearchUser*)object;

				if ( HubSearchClient->m_pClient->IsUserOnline( msg->m_sNick ) == TRUE )
				{
					CMessageSearchResultUser * MessageSearchResultUser = new CMessageSearchResultUser();

					MessageSearchResultUser->m_eType    = DC_MESSAGE_SEARCHRESULT_USER;
					MessageSearchResultUser->m_sHubName = HubSearchClient->m_pClient->GetHubName();
					MessageSearchResultUser->m_sNick    = msg->m_sNick;

					SendSearchResult(MessageSearchResultUser);
				}

				break;
			}

			case DC_MESSAGE_SEARCH_FILE:
			{
				CMessageSearchFile * msg = (CMessageSearchFile*)object;

				search = GetSearchString( msg, m_bMultiSearch );

				if ( search != "" )
					HubSearchClient->m_pClient->SendString(search);

				// update current object
				HubSearchClient->m_pCurrentSearch = object;
				// update timeout
				HubSearchClient->m_tSearchTimeout = time(0);

				res = TRUE;

				break;
			}

			default:
			{
				break;
			}
		}
	}

	return res;
}

/** */
int CHubSearch::DC_ClientCallBack( CObject * Client, CObject * Object )
{
	_pHubSearch->ClientThread.Lock();

	CClient * c = (CClient*)Client;
	CObject * o = Object;

	if ( (c == 0) || (o == 0) )
	{
		_pHubSearch->ClientThread.UnLock();
		return -1;
	}

	CDCMessage *DCMsg;

	DCMsg = (CDCMessage*) o;

	switch ( DCMsg->m_eType )
	{
		case DC_MESSAGE_CONNECTION_STATE:
		{
			CMessageConnectionState *msg = (CMessageConnectionState*) o;

			switch(msg->m_eState)
			{
				case estCONNECTED:
				{
					break;
				}

				case estSOCKETERROR:
				{
					_pHubSearch->SendDebug( "SocketError on " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) + " [" + msg->m_sMessage + "]");
					m_nError++;
					break;
				}

				case estDISCONNECTED:
				{
					_pHubSearch->RemoveClient(c);
				}

				default:
				{
					break;
				}
			}

			break;
		}

		case DC_MESSAGE_VALIDATEDENIDE:
		{
			_pHubSearch->SendDebug( "Validate denide " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) );
			c->Disconnect(TRUE);
			break;
		}

		case DC_MESSAGE_HUBISFULL:
		{
			_pHubSearch->SendDebug( "Hub is full on " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) );
			c->Disconnect(TRUE);
			break;
		}

		case DC_MESSAGE_FORCEMOVE:
		{
			_pHubSearch->SendDebug( "Force move on " + c->GetIP() + CString(":") + CString().setNum(c->GetPort()) );
			c->Disconnect(TRUE);
			break;
		}

		case DC_MESSAGE_LOCK:
		case DC_MESSAGE_HUBNAME:
		case DC_MESSAGE_CHAT:
		case DC_MESSAGE_MYNICK:
		case DC_MESSAGE_PRIVATECHAT:
		case DC_MESSAGE_SEARCH:
		case DC_MESSAGE_HELLO:
		case DC_MESSAGE_QUIT:
		case DC_MESSAGE_GETINFO:
		{
			break;
		}

		case DC_MESSAGE_MYINFO:
		{
			CString s;
			CMessageMyInfo * MessageMyInfo = (CMessageMyInfo *)o;

			if ( MessageMyInfo->m_sNick == CConfig::Instance()->GetSearchNick() )
			{
				if ( _pHubSearch->m_bHandleUserList == FALSE )
				{
					// enable search
					EnableSearch(c);
					// send search
					//_pHubSearch->SendSearch(c);

					// remove the client only for active searches
					//if ( CConfig::Instance()->GetMode() != ecmPassive )
					//	_pHubSearch->RemoveClient(c);
				}
			}
			break;
		}

		case DC_MESSAGE_NICKLIST:
		{
			if ( _pHubSearch->m_bHandleUserList == TRUE )
			{
				// enable search
				EnableSearch(c);

				// send search
				//_pHubSearch->SendSearch(c);

				// remove the client only for active searches
				//if ( CConfig::Instance()->GetMode() != ecmPassive )
				//	_pHubSearch->RemoveClient(c);
			}
			break;
		}

		case DC_MESSAGE_SEARCHRESULT:
		{
			if ( CConfig::Instance()->GetMode() == ecmPASSIVE )
			{
				_pHubSearch->SendSearchResult(o);
				o = 0;
			}
			break;
		}

		default:
		{
			printf("callback: %d\n",DCMsg->m_eType);
			break;
		}
	}

	if ( o )
		delete o;

	_pHubSearch->ClientThread.UnLock();

	return 0;
}
