/***************************************************************************
                           casyncdns.cpp  -  description
                             -------------------
    begin                : Mon Jun 10 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef WIN32
#include <unistd.h>
#include <netdb.h>
#endif

#include <string.h>

#include "clist.h"
#include "cstringlist.h"
#include "csocket.h"
#include "casyncdns.h"

/** */
CAsyncDns::CAsyncDns()
{
	m_pHostCache = new CStringList(4);
	m_pHostList  = new CStringList(4);

	Start();

	SetInstance(this);
}

/** */
CAsyncDns::~CAsyncDns()
{
	Stop();

	Lock();

	if ( m_pHostCache )
	{
		delete m_pHostCache;
		m_pHostCache = 0;
	}

	if ( m_pHostList )
	{
		delete m_pHostList;
		m_pHostList = 0;
	}

	UnLock();
}

/** */
eAsyncDns CAsyncDns::GetHostByName( CString stringHost, struct in_addr * sin_addr, int * numerr )
{
	if ( TryLock() == FALSE )
	{
		return eadAGAIN;
	}

	CAsyncDnsEntry * AsyncDnsEntry;
	eAsyncDns err = eadERROR;

	if ( m_pHostCache && m_pHostList && (stringHost != "") && sin_addr )
	{
		if ( m_pHostCache->Get( stringHost, (CObject*&)AsyncDnsEntry ) != -1 )
		{
			memcpy( sin_addr, &AsyncDnsEntry->m_sin_addr, sizeof(struct in_addr) );
			m_pHostCache->Del(stringHost);
			err = eadSUCCESS;
		}

		if ( err == eadERROR )
		{
			if ( m_pHostList->Get( stringHost, (CObject*&)AsyncDnsEntry ) == -1 )
			{
				AsyncDnsEntry = new CAsyncDnsEntry();

				AsyncDnsEntry->m_sHost  = stringHost;
				AsyncDnsEntry->m_bError = FALSE;

				m_pHostList->Add( stringHost, AsyncDnsEntry );

				err = eadAGAIN;
			}
			else if ( AsyncDnsEntry->m_bError == FALSE )
			{
				err = eadAGAIN;
			}
			else if ( numerr != 0 )
			{
				*numerr = AsyncDnsEntry->m_nError;
			}
		}
	}

	UnLock();

	return err;
}

/** */
void CAsyncDns::Thread( CObject * )
{
	CAsyncDnsEntry * AsyncDnsEntry = 0;
	CString host = "";
	struct hostent *hp = 0;

	Lock();

	while ( m_pHostList->Next( (CObject*&)AsyncDnsEntry ) != 0 )
	{
		// no error
		if ( AsyncDnsEntry->m_bError == FALSE )
		{
			host = AsyncDnsEntry->m_sHost;
			break;
		}
		else if ( (time(0)-AsyncDnsEntry->m_tTimeout) >= 60 )
		{
			m_pHostList->Del(AsyncDnsEntry->m_sHost);
			break;
		}
	}

	UnLock();

	if ( host != "" )
	{
		CSocket::GetHostByName( host.Data(), hp );

		Lock();

		if ( m_pHostList->Get( host, (CObject*&)AsyncDnsEntry ) == 0 )
		{
			if ( hp == 0 )
			{
				AsyncDnsEntry->m_bError   = TRUE;
				AsyncDnsEntry->m_tTimeout = time(0);
#ifdef WIN32
				AsyncDnsEntry->m_nError = WSAGetLastError();
#else
				AsyncDnsEntry->m_nError = h_errno;
#endif
			}
			else
			{
				AsyncDnsEntry = new CAsyncDnsEntry();

				AsyncDnsEntry->m_sHost    = host;
				AsyncDnsEntry->m_bError   = FALSE;
				AsyncDnsEntry->m_tTimeout = time(0);
				memcpy( &AsyncDnsEntry->m_sin_addr, hp->h_addr, sizeof(struct in_addr) );

				m_pHostList->Del(host);
				m_pHostCache->Add(host,AsyncDnsEntry);
			}
		}

		UnLock();
	}

	NanoSleep(50);
}
