/***************************************************************************
                           cfile.cpp  -  description
                             -------------------
    begin                : Wed Apr 3 2003
    copyright            : (C) 2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#ifndef WIN32
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#endif

#include "cfile.h"

/** */
CFile::CFile()
{
	m_nFD = -1;
}

/** */
CFile::~CFile()
{
	Close();
}

/** */
bool CFile::Open( CString filename, int mode, int acc )
{
	bool res = FALSE;
	int m = 0, a = 0;

	if ( m_nFD != -1 )
	{
		return res;
	}

#ifdef WIN32
	if ( (mode & IO_RAW) != 0 )
		m |= _O_BINARY;
	if ( (mode & IO_READONLY) != 0 )
		m |= _O_RDONLY;
	if ( (mode & IO_WRITEONLY) != 0 )
		m |= _O_WRONLY;
	if ( (mode & IO_READWRITE) != 0 )
		m |= _O_RDWR;
	if ( (mode & IO_APPEND) != 0 )
		m |= _O_APPEND;
	if ( (mode & IO_TRUNCATE) != 0 )
		m |= _O_TRUNC;
	if ( (mode & IO_CREAT) != 0 )
		m |= _O_CREAT;
#else
	if ( (mode & IO_RAW) != 0 )
		m |= 0;
	if ( (mode & IO_READONLY) != 0 )
		m |= O_RDONLY;
	if ( (mode & IO_WRITEONLY) != 0 )
		m |= O_WRONLY;
	if ( (mode & IO_READWRITE) != 0 )
		m |= O_RDWR;
	if ( (mode & IO_APPEND) != 0 )
		m |= O_APPEND;
	if ( (mode & IO_TRUNCATE) != 0 )
		m |= O_TRUNC;
	if ( (mode & IO_CREAT) != 0 )
		m |= O_CREAT;
#endif

#ifdef WIN32
	if ( (acc & MO_IRWXU) != 0 )
		a |= _S_IREAD | _S_IWRITE;
	if ( (acc & MO_IRUSR) != 0 )
		a |= _S_IREAD;
	if ( (acc & MO_IWUSR) != 0 )
		a |= _S_IWRITE;
	if ( (acc & MO_IRWXG) != 0 )
		a |= _S_IREAD | _S_IWRITE;
	if ( (acc & MO_IRGRP) != 0 )
		a |= _S_IREAD;
	if ( (acc & MO_IWGRP) != 0 )
		a |= _S_IWRITE;
	if ( (acc & MO_IRWXO) != 0 )
		a |= _S_IREAD | _S_IWRITE;
	if ( (acc & MO_IRGRP) != 0 )
		a |= _S_IREAD;
	if ( (acc & MO_IWOTH) != 0 )
		a |= _S_IWRITE;
#else
	if ( (acc & MO_IRWXU) != 0 )
		a |= S_IRWXU;
	if ( (acc & MO_IRUSR) != 0 )
		a |= S_IRUSR;
	if ( (acc & MO_IWUSR) != 0 )
		a |= S_IWUSR;
	if ( (acc & MO_IXUSR) != 0 )
		a |= S_IXUSR;
	if ( (acc & MO_IRWXG) != 0 )
		a |= S_IRWXG;
	if ( (acc & MO_IRGRP) != 0 )
		a |= S_IRGRP;
	if ( (acc & MO_IWGRP) != 0 )
		a |= S_IWGRP;
	if ( (acc & MO_IXGRP) != 0 )
		a |= S_IXGRP;
	if ( (acc & MO_IRWXO) != 0 )
		a |= S_IRWXO;
	if ( (acc & MO_IRGRP) != 0 )
		a |= S_IROTH;
	if ( (acc & MO_IWOTH) != 0 )
		a |= S_IWOTH;
	if ( (acc & MO_IXOTH) != 0 )
		a |= S_IXOTH;
#endif

#ifdef WIN32
	m_nFD = _open( filename.Data(), m, a );
#else
	m_nFD = open( filename.Data(), m, a );
#endif

	if ( m_nFD != -1 )
	{
		res = TRUE;
	}

	return res;
}

/** */
void CFile::Close()
{
	if ( m_nFD != -1 )
	{
#ifdef WIN32
		_close(m_nFD);
#else
		close(m_nFD);
#endif
		m_nFD = -1;
	}
}

/** */
long CFile::Write( const char * buf, long count )
{
	long i = -1;

	if ( (m_nFD == -1) || (count <= 0) || (!buf))
	{
		return i;
	}

#ifdef WIN32
	i = _write( m_nFD, buf, count );
#else
	i = write( m_nFD, buf, count );

#endif

	return i;
}

/** */
long CFile::Read( char * buf, long count )
{
	long i = -1;

	if ( (m_nFD == -1) || (count <= 0) || (!buf))
	{
		return i;
	}

#ifdef WIN32
	i = _read( m_nFD, buf, count );
#else
	i = read( m_nFD, buf, count );
#endif

	return i;
}

/** */
bool CFile::Seek( ulonglong offset, int origin )
{
	bool res = FALSE;

	if ( m_nFD == -1 )
	{
		return res;
	}

#ifdef WIN32
	if ( _lseeki64( m_nFD, offset, origin ) == offset )
#else
	if ( lseek( m_nFD, offset, origin ) == offset )
#endif
	{
		res = TRUE;
	}

	return res;
}
