/***************************************************************************
                          cservermanager.cpp  -  description
                             -------------------
    begin                : Don Mai 16 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <dclib/cclient.h>
#include <dclib/cconfig.h>
#include <dclib/core/cmanager.h>
#include <dclib/cdownloadmanager.h>
#include <dclib/chubsearch.h>
#include <dclib/dcobject.h>
#include <dclib/core/cbytearray.h>
#include <dclib/chttp.h>
#include <dclib/core/cbz.h>
#include <dclib/cfilemanager.h>

#ifndef WIN32
#include <unistd.h>
#include <stdlib.h>
#endif

#include "cservermanager.h"

/** */
CServerManager::CServerManager()
{
	m_pClientList = new CThreadList<CClient>();

	m_pCallback = new CCallback<CServerManager>( this, &CServerManager::Callback );
	CManager::Instance()->Add( m_pCallback );

	if ( CConfig::Instance()->GetReloadHubListTime() != 0 )
	{
		m_nReloadHubListTimeout = time(0) + CConfig::Instance()->GetReloadHubListTime()*60*60;
	}
	else
	{
		m_nReloadHubListTimeout = 0;
	}

	// get hublist stuff
	m_pHttp           = 0;
	m_pHubListUrlList = 0;
	m_pHubListUrl     = 0;
	m_pHubListData    = 0;

	m_bGetHubListDone = FALSE;

	SetInstance(this);
}

/** */
CServerManager::~CServerManager()
{
	m_Thread.Stop();

	CManager::Instance()->Remove( m_pCallback );
	if (m_pCallback)
	{
		delete m_pCallback;
		m_pCallback = 0;
	}
	
	m_Thread.Lock();
	m_pClientList->Lock();

	delete m_pClientList;
	m_pClientList = 0;

	m_Thread.UnLock();
}

/** thread callbackfunction */
int CServerManager::Callback( CObject *, CObject * )
{
	CClient * client = 0;

	m_Thread.Lock();

	if ( m_pClientList )
	{
		while( (client=m_pClientList->Next(client)) != 0 )
		{
			client->Thread(0);
		}
	}

	// check reload hublist timeout
	if ( CConfig::Instance()->GetReloadHubListTime() != 0 )
	{
		if ( m_nReloadHubListTimeout != 0 )
		{
			if ( time(0) >= m_nReloadHubListTimeout )
			{
				GetPublicHubList();

				m_nReloadHubListTimeout = time(0) + CConfig::Instance()->GetReloadHubListTime()*60*60;
			}
		}
		else // change from config
		{
			m_nReloadHubListTimeout = time(0) + CConfig::Instance()->GetReloadHubListTime()*60*60;
		}
	}
	else
	{
		m_nReloadHubListTimeout = 0;
	}

	// cleanup all object
	if ( m_bGetHubListDone == TRUE )
	{
		if ( m_pHttp )
		{
			delete m_pHttp;
			m_pHttp = 0;
		}

		if ( m_pHubListUrlList )
		{
			delete m_pHubListUrlList;
			m_pHubListUrlList = 0;
		}

		if ( m_pHubListData )
		{
			delete m_pHubListData;
			m_pHubListData = 0;
		}

		m_pHubListUrl = 0;

		DCMessageGetHubList * msg = new DCMessageGetHubList();

		msg->m_bRun = FALSE;

		if ( DC_CallBack(msg) == -1 )
		{
			delete msg;
		}

		m_bGetHubListDone = FALSE;
	}

	m_Thread.UnLock();

	return 0;
}

/** http callback function */
int CServerManager::HttpCallBack( CObject *, CObject * object )
{
	CByteArray in,out;
	CBZ bz2;

	switch(((CDCMessage*)object)->m_eType)
	{
		case DC_MESSAGE_CONNECTION_STATE:
		{
			CMessageConnectionState *msg = (CMessageConnectionState*)object;

			if ( msg->m_eState == estDISCONNECTED )
			{
				if ( m_pHttp->GetData(&in) == TRUE )
				{
					if ( m_pHttp->GetUrl().Find(".bz2") != -1 )
					{
						if ( bz2.Decompress( &in, &out ) == TRUE )
						{
							m_pHubListData->Append(out.Data(),out.Size());
							m_pHubListData->Append("\xD\xA",2);
						}
						else
						{
							printf("bz2 decompress failed\n");
						}
					}
					else
					{
						m_pHubListData->Append(in.Data(),in.Size());
						m_pHubListData->Append("\xD\xA",2);
					}

				}

				if ( NextHubListUrl() == FALSE )
				{
					// parse public hub list in own thread
					CCallback<CServerManager> * c = new CCallback<CServerManager>( this, &CServerManager::ParsePublicHubList );
					m_Thread.SetThreadCallBackFunction(c);
					m_Thread.Start();
				}
			}

			break;
		}

		case DC_MESSAGE_TRANSFER:
		{
			if ( DC_CallBack(object) != -1 )
			{
				object = 0;
			}

			break;
		}

		default:
		{
			break;
		}
	}

	if ( object )
	{
		delete object;
	}

	return 0;
}

/** */
bool CServerManager::GetPublicHubList()
{
	bool res = FALSE;

	if ( m_pHttp != 0 )
	{
		
		// get hublist allready running
		return res;
	}

	m_pHubListUrlList = new CList<DCConfigHubListUrl>();

	// get all hublist urls
	CConfig::Instance()->GetHubListUrlList(m_pHubListUrlList);

	// check hublist url count
	if ( m_pHubListUrlList->Count() == 0 )
	{
		delete m_pHubListUrlList;
		m_pHubListUrlList = 0;
		return res;
	}

	// init first url
	m_pHubListUrl = 0;

	m_pHubListData = new CByteArray();
	m_pHttp = new CHttp();
	m_pHttp->SetCallBackFunction( new CCallback<CServerManager>( this, &CServerManager::HttpCallBack ) );

	res = NextHubListUrl();

	if ( !res )
	{
		m_bGetHubListDone = TRUE;
	}
	else
	{
		DCMessageGetHubList * msg = new DCMessageGetHubList();

		msg->m_bRun = TRUE;

		if ( DC_CallBack(msg) == -1 )
		{
			delete msg;
		}
	}

	return res;
}

/** */
int CServerManager::ParsePublicHubList( CObject * , CObject * )
{
	CString line;
	CString s;
	CString s1,s2,s3,s4;
	long i=0,i1=0;

	if ( m_pHubListData->Size() == 0 )
	{
		m_Thread.Stop();
		m_Thread.SetThreadCallBackFunction(0);
		m_bGetHubListDone = TRUE;

		return 0;
	}

	s.Set((const char*)m_pHubListData->Data(),m_pHubListData->Size());

	while( (i = s.Find(0x0d,i)) != -1 )
	{
		line = s.Mid(i1,i-i1);

		if ( !line.IsEmpty() )
		{
			s1 = line.Section( '|', 0, 0 );
			s2 = line.Section( '|', 1, 1 );
			s3 = line.Section( '|', 2, 2 );
			s4 = line.Section( '|', 3, 3 );

			// replace all spaces
			s2 = s2.Replace(" ","");

			CConfig::Instance()->AddPublicHub( s1, s2, s3, s4 );
		}

		i1 = i+2;
		i += 2;
	}

	// store the list
	if ( CConfig::Instance()->GetHubListStoreLocal() == TRUE )
	{
		CConfig::Instance()->SaveDCHub();
	}

	m_Thread.Stop();
	m_Thread.SetThreadCallBackFunction(0);
	m_bGetHubListDone = TRUE;

	return 0;
}

/** */
bool CServerManager::NextHubListUrl()
{
	bool res = FALSE;

	while( (m_pHubListUrl=m_pHubListUrlList->Next(m_pHubListUrl)) != 0 )
	{
		if ( m_pHubListUrl->bEnabled == TRUE )
		{
			m_pHttp->GetUrl(m_pHubListUrl->sUrl);

			res = TRUE;
			break;
		}
	}

	return res;
}

/** */
void CServerManager::ConnectClient( CString hubname, CString server )
{
	DCMessageConnectClient *msg = new DCMessageConnectClient();

	msg->m_sHubName = hubname;
	msg->m_sHubHost = server;

	if ( DC_CallBack(msg) == -1 )
	{
		delete msg;

		Connect( hubname, server, NULL );
	}
}

/** */
void CServerManager::Connect( CString hubname, CString server, CClient * client )
{
	if ( !client )
	{
		if ( server == "" )
		{
			return;
		}

		if ( hubname == "" )
		{
			hubname = server;
		}


		m_Thread.Lock();

		client = (CClient*)GetHub(hubname);

		m_Thread.UnLock();

		if ( client )
		{
			return;
		}

		client = new CClient();
	}

	client->SetHubName(hubname);

	AddHub(client);

	client->SetNick(CConfig::Instance()->GetNick( hubname ));
	client->SetComment(CConfig::Instance()->GetDescription());
	client->SetConnectionType(CConfig::Instance()->GetSpeed());
	client->SetEMail(CConfig::Instance()->GetEMail());
	client->SetVersion(CString("DCGUI-") + VERSION);
	client->SetShareSize(CString().setNum(CFileManager::Instance()->GetShareSize()));
	client->SetMode(CConfig::Instance()->GetMode());
	client->SetAwayMode(CConfig::Instance()->GetAwayMode());
	client->Connect(server);
}

/** */
void CServerManager::AddHub( CClient * client )
{
	m_Thread.Lock();

	if ( m_pClientList )
	{
		m_pClientList->Lock();
		m_pClientList->Add(client);
		m_pClientList->UnLock();
	}

	m_Thread.UnLock();
}

/** */
void CServerManager::RemoveHub( CClient * client )
{
	m_Thread.Lock();

	if ( m_pClientList )
	{
		m_pClientList->Lock();
		client->SetCallBackFunction(0);
		m_pClientList->Remove(client);
		m_pClientList->UnLock();
	}

	m_Thread.UnLock();
}

/** */
int CServerManager::SM_ClientCallBack( CObject * /*Sender*/, CObject * /*Object*/ )
{
//	Transfer->SetCallBackFunction( new CCallback<CDownloadManager>( pDownloadManager, &CDownloadManager::DM_TransferCallBack ) );
	return -1;
}

/** */
long CServerManager::GetConnectedHubCount( bool admin )
{
	long i;
	CClient * client;

	if ( !m_pClientList )
	{
		return 0;
	}

	m_pClientList->Lock();

	i = 0;
	client = 0;

	while( (client=m_pClientList->Next(client)) != 0 )
	{
		if ( (client->IsHandshake() == TRUE) )
			continue;

		if ( (admin == TRUE) && (client->IsAdmin() == TRUE) )
			continue;

		i++;
	}

	m_pClientList->UnLock();

	return i;
}

/** */
CClient * CServerManager::GetHub( CString hubname )
{
	CClient* client = 0;

	if ( !m_pClientList )
	{
		return client;
	}

	m_pClientList->Lock();

	while( (client=m_pClientList->Next(client)) != 0 )
	{
		if ( client->GetHubName() == hubname )
		{
			break;
		}
	}

	m_pClientList->UnLock();

	return client;
}

/** */
CStringList * CServerManager::GetPublicHubServerList()
{
	DCConfigHubItem * hubitem;
	CStringList * StringList = 0;
	CObject * object;

	CList<DCConfigHubItem> List;

	CConfig::Instance()->GetPublicHubList(&List);

	if ( List.Count() > 0 )
	{
		hubitem = 0;
		StringList = new CStringList();

		while( (hubitem=List.Next(hubitem)) != 0 )
		{
			if ( StringList->Get( hubitem->sHost, (CObject *&)object ) != 0 )
			{
				StringList->Add(hubitem->sHost, new CString(hubitem->sHost));
			}
		}
	}

	return StringList;
}

/** */
CStringList * CServerManager::GetBookmarkHubServerList()
{
	DCConfigHubItem * hubitem;
	CStringList * StringList = 0;
	CObject * object;

	CList<DCConfigHubItem> List;

	CConfig::Instance()->GetBookmarkHubList(&List);

	if ( List.Count() > 0 )
	{
		hubitem = 0;
		StringList = new CStringList();

		while( (hubitem=List.Next(hubitem)) != 0 )
		{
			if ( StringList->Get( hubitem->sHost, (CObject *&)object ) != 0 )
			{
				StringList->Add(hubitem->sHost, new CString(hubitem->sHost));
			}
		}
	}

	return StringList;
}

/** */
CStringList * CServerManager::GetConnectedHubServerList()
{
	CClient* client;
	CStringList * StringList;

	if ( !m_pClientList )
	{
		return 0;
	}

	m_pClientList->Lock();

 	StringList = new CStringList();

	if ( m_pClientList->Count() > 0 )
	{
		client = 0;
		while( (client=m_pClientList->Next(client)) != 0 )
		{
			if ( client->IsHandshake() == TRUE )
			{
				continue;
			}

			StringList->Add("",new CString(client->GetHubName()));
		}
	}

	m_pClientList->UnLock();

	return StringList;
}

/** send message to all connected servers */
int CServerManager::SendStringToConnectedServers( CString s, CString hubname )
{
	int i;
	CClient* client;

	if ( !m_pClientList )
	{
		return 0;
	}

	m_pClientList->Lock();

	i = 0;

	if ( s != "" )
	{
		client = 0;

		while( (client=m_pClientList->Next(client)) != 0 )
		{
			if ( client->IsHandshake() == TRUE )
			{
				continue;
			}

			if ( hubname != "" )
			{
				if ( hubname == client->GetHubName() )
				{
					client->SendString(s);
					i++;
					break;
				}
				else
				{
					continue;
				}
			}
			else
			{
				client->SendString(s);
			}

			i++;
		}
	}

	m_pClientList->UnLock();

	return i;
}

/** send a search string (extracted from message) to all connected servers */
int CServerManager::SendStringToConnectedServers( CMessageSearchFile *sf )
{
	CString s;
	int i;
	CClient* client;

	if ( !m_pClientList )
	{
		return 0;
	}

	m_pClientList->Lock();

	i = 0;
	client = 0;

	while( (client=m_pClientList->Next(client)) != 0 )
	{
		sf->m_sSource = CConfig::Instance()->GetNick( client->GetHubName() );

		s = CHubSearch::Instance()->GetSearchString(sf);

		if ( s == "" )
		{
			break;
		}

		if ( client->IsHandshake() == TRUE )
		{
			continue;
		}

		client->SendString(s);

		i++;
	}

	m_pClientList->UnLock();

	return i;
}

/** send myinfo to all connected servers */
int CServerManager::SendMyInfoToConnectedServers()
{
	CClient* client;
	CString description,speed,email,share;
	eUserAwayMode awaymode;
	int i;

	if ( !m_pClientList ||
	     !CFileManager::Instance() ||
	     !CConfig::Instance() )
	{
		return 0;
	}

	description = CConfig::Instance()->GetDescription();
	speed       = CConfig::Instance()->GetSpeed();
	email       = CConfig::Instance()->GetEMail();
	share       = CString().setNum(CFileManager::Instance()->GetShareSize());
	awaymode    = CConfig::Instance()->GetAwayMode();

	m_pClientList->Lock();

	i = 0;
	client = 0;

	while( (client=m_pClientList->Next(client)) != 0 )
	{
		if ( client->IsHandshake() == TRUE )
		{
			continue;
		}

		// update internal values
		client->SetComment(description);
		client->SetConnectionType(speed);
		client->SetEMail(email);
		client->SetShareSize(share);
		client->SetAwayMode(awaymode);
		i++;
	}
	m_pClientList->UnLock();

	return i;
}

/** */
int CServerManager::SendConnectionRequest( CString nick, CString hubname )
{
	int res;
	CClient* client;

	if ( !m_pClientList )
	{
		return 0;
	}

	m_pClientList->Lock();

	res = -3;
	client = 0;

	while( (client=m_pClientList->Next(client)) != 0 )
	{
		if ( client->GetHubName() == hubname )
		{
			if ( client->IsHandshake() == TRUE )
			{
				res = -2;
				break;
			}

			if ( client->IsUserOnline( nick ) == FALSE )
			{
				res = -1;
				break;
			}

			if ( client->GetMode() == ecmPASSIVE )
			{
				// send connection request
				if ( client->SendRevConnectToMe( client->GetNick(), nick ) == 0 )
				{
					res = 0;
				}
				else
				{
					res = -4;
				}
			}
			else if ( client->GetMode() == ecmACTIVE )
			{
				CString s = CConfig::Instance()->GetTCPHostString();

				if ( s != "" )
				{
					// send connection request
					CDownloadManager::Instance()->DLM_AddTransferRequest( nick, "", hubname, client->GetHost() );

					if ( client->SendConnectToMe( nick, s ) == 0 )
					{
						res = 0;
					}
					else
					{
						res = -4;
					}
				}
				else
				{
					res = -4;
				}
			}

			break;
		}
	}
	m_pClientList->UnLock();

	return res;
}

/** */
bool CServerManager::IsUserOnline( CString nick, CString hubname, CList<DCHubObject> * list )
{
	bool res;
	CClient* client;
	DCHubObject * HubObject;

	if ( !m_pClientList )
	{
		return FALSE;
	}

	m_pClientList->Lock();

	res = FALSE;
	client = 0;

	while( (client=m_pClientList->Next(client)) != 0 )
	{
		if ( client->IsHandshake() == FALSE )
		{
			if ( hubname == "" )
			{
				if ( client->IsUserOnline( nick ) == TRUE )
				{
					if ( list )
					{
						HubObject = new DCHubObject();
						HubObject->m_sHubName = client->GetHubName();
						HubObject->m_sHubHost = client->GetHost();
						list->Add(HubObject);
					}
					res = TRUE;
				}
			}
			else if ( client->GetHubName() == hubname )
			{
				if ( client->IsUserOnline( nick ) == TRUE )
					res = TRUE;
				break;
			}
		}
	}
	m_pClientList->UnLock();

	return res;
}

/** */
eHubState CServerManager::IsHubOnline( CString hubname )
{
	eHubState res;
	CClient* client;

	if ( !m_pClientList )
	{
		return ehsNONE;
	}

	m_pClientList->Lock();

	res = ehsNONE;
	client = 0;

	while( (client=m_pClientList->Next(client)) != 0 )
	{
		if ( client->GetHubName() == hubname )
		{
			if ( client->IsHandshake() == FALSE )
			{
				res = ehsONLINE;
			}
			else
			{
				res = ehsOFFLINE;
			}

			break;
		}
	}
	m_pClientList->UnLock();

	return res;
}

/** */
bool CServerManager::IsAdmin( CString hubname, CString nick )
{
	bool res;
	CClient* client;

	if ( !m_pClientList )
	{
		return FALSE;
	}

	m_pClientList->Lock();

	res = FALSE;
	client = 0;

	while( (client=m_pClientList->Next(client)) != 0 )
	{
		if ( client->GetHubName() == hubname )
		{
			if ( client->IsHandshake() == FALSE )
			{
				if ( nick == "" )
					res = client->IsAdmin();
				else
					res = client->IsAdmin(nick);
			}

			break;
		}
	}
	m_pClientList->UnLock();

	return res;
}

/** */
CString CServerManager::GetHubHost( CString hubname )
{
	CString res;
	CClient* client;

	if ( !m_pClientList )
	{
		return "";
	}

	m_pClientList->Lock();

 	res = "";
	client = 0;

	while( (client=m_pClientList->Next(client)) != 0 )
	{
		if ( client->GetHubName() == hubname )
		{
			res = client->GetIP() + ":" + CString().setNum(client->GetPort());
			break;
		}
	}
	m_pClientList->UnLock();

	return res;
}
